package tech.nevets.modbot;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import tech.nevets.modbot.api.IBotModule;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;

public class ModuleLoader {
    private static final Logger LOGGER = LoggerFactory.getLogger(ModuleLoader.class);

    public static List<Path> getModules() {
        File modulesPath = new File("./modules/");
        if (!modulesPath.exists()) {
            modulesPath.mkdirs();
        }

        List<Path> modules = new ArrayList<>();
        try {
            Files.walk(Paths.get("./modules/"))
                    .filter(Files::isRegularFile)
                    .forEach(modules::add);
        } catch (IOException e) {
            LOGGER.error("Unable to get module list");
            e.printStackTrace();
        }

        return modules;
    }

    //TODO: make this dynamically wrap all interface modules into abstract modules
    //TODO: handle both types of modules

    public static List<IBotModule> loadModules() {
        List<Path> modulePaths = getModules();
        List<IBotModule> modules = new ArrayList<>();

        for (Path module : modulePaths) {
            JarInputStream jarStream;
            Manifest mf;
            Attributes attributes;
            String mainClassName;

            try {
                jarStream = new JarInputStream(new FileInputStream(String.valueOf(module)));
                mf = jarStream.getManifest();
                attributes = mf.getMainAttributes();
                mainClassName = attributes.getValue("Main-Class");
            } catch (final Exception e) {
                LOGGER.error("Error while reading module attributes from module: " + module.toString());
                e.printStackTrace();
                continue;
            }

            if (mainClassName == null) {
                LOGGER.error("Unable to get mainClassName from module: " + module.toString());
                continue;
            }

            File file = module.toFile();
            URL[] url;
            try {
                url = new URL[]{file.toURI().toURL()};
            } catch (MalformedURLException e) {
                LOGGER.error("Error occurred while getting file URL");
                e.printStackTrace();
                continue;
            }
            URLClassLoader urlClassLoader = URLClassLoader.newInstance(url);

            Class<?> mainClass;
            try {
                mainClass = Class.forName(mainClassName, true, urlClassLoader);
            } catch (ClassNotFoundException e) {
                LOGGER.error("Error getting class from classLoader, Class Not Found!");
                e.printStackTrace();
                continue;
            }

            String moduleType = attributes.getValue("Module-Type");
            if (moduleType.equalsIgnoreCase("inherit") || moduleType.equalsIgnoreCase("i")) {

            } else if (moduleType.equalsIgnoreCase("extend") || moduleType.equalsIgnoreCase("e")) {

            } else {
                LOGGER.error("Error loading module: Module " + module.toString() + " has an invalid Module-Type!");
                continue;
            }

            IBotModule newModule;
            try {
                newModule = (IBotModule) mainClass.getDeclaredConstructor().newInstance();
            } catch (NoSuchMethodException | InstantiationException | IllegalAccessException | InvocationTargetException e) {
                LOGGER.error("Error while creating instance of module " + module.toString());
                e.printStackTrace();
                continue;
            }
            newModule.onPreEnable();
            modules.add(newModule);

            try {
                jarStream.close();
                if (urlClassLoader != null) {
                    urlClassLoader.close();
                }
            } catch (IOException e) {
                LOGGER.error("Error closing jarStream or classLoader");
                e.printStackTrace();
            }
        }

        LOGGER.info("Loaded module: " + modules.toArray().toString());
        return modules;
    }
}
